//*****************************************************************************
//* =====================
//*  AHSpring
//* =====================
//*  
//* (Description)
//*    e+e- --> AH generator
//*
//* (Update Record)
//*    2021/10/08  K.Fujii	Original version.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "AHSpring.h"

#include <sstream>
#include <iomanip>
//#define __DEBUG__
//#define __PHASESPACE__

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

ClassImp(AHSpring)
ClassImp(AHSpringBuf)
ClassImp(AHBases)

//-----------------------------------------------------------------------------
// ==============================
//  class AHSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
AHSpring::AHSpring(const char    *name,
                   const char    *title,
                         AHBases *bases)
         : JSFSpring(name, title, bases)
{
  fEventBuf = new AHSpringBuf("AHSpringBuf",
                              "AHSpring event buffer",
                              this);
  if (!bases) { 
    SetBases(new AHBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
AHSpring::~AHSpring()
{
  delete fEventBuf;
  delete GetBases();
}


//-----------------------------------------------------------------------------
// ==============================
//  class AHSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t AHSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  AHBases    *bases   = static_cast<AHBases*>(
		        static_cast<AHSpring*>(Module())->GetBases());

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  const Int_t kNparton = 2;
  ANL4DVector pv[kNparton];
  pv[0] = bases->fP[0];  // photon
  pv[1] = bases->fP[1];  // higgs

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fQ2AH         = fEcmIP*fEcmIP;
  fEta          = bases->GetEta();
  fPhi          = bases->GetPhi();
  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);
  for (Int_t i=0; i<kNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  fNparton = kNparton;

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [kNparton];
    for (Int_t i=0; i<kNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<kNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }

  Int_t    ida     = 22;                          // PDG code for photon
  Int_t    idh     = 25;                          // PDG code for H
  Int_t    hela    = bases->fHelFinal[0];         // photon helicity
  Int_t    helh    = bases->fHelFinal[1];         // higgs helicity
  Double_t rq2h    = pv[1].Mag();

#ifdef __DEBUG__
  cerr << " -------------------------- " << endl;
  cerr << " 1 pid=" << ida << " m=" << 0.  << " Q=" << 0.
       << " pv=(" << (*qp[0])(0) << ", "
                  << (*qp[0])(1) << ", "
                  << (*qp[0])(2) << ", "
                  << (*qp[0])(3) << ") " << endl;
  cerr << " 2 pid=" << idh << " m=" << rq2h << " Q=" << 0.
       << " pv=(" << (*qp[1])(0) << ", "
                  << (*qp[1])(1) << ", "
                  << (*qp[1])(2) << ", "
                  << (*qp[1])(3) << ") " << endl;
  TVector qcm(4);
  qcm = *qp[0] + *qp[1];
  cerr << " pcm=(" << qcm[0] << ", "
                   << qcm[1] << ", "
                   << qcm[2] << ", "
                   << qcm[3] << ") " << endl;
#endif

  //                              No. PID  Mass  Charge   pv   Nd 1st Mom  hel  col shower
  new (partons[0]) JSFSpringParton(1, ida,   0.,    0., *qp[0], 0, 0,  0, hela,   0,     0);
  new (partons[1]) JSFSpringParton(2, idh, rq2h,    0., *qp[1], 0, 0,  0, helh,   0,     0);
  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class AHBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
AHBases::AHBases(const char *name, const char *title)
         : JSFBases   (name, title), 
           fMass      ( 125.),
           fEcmInit   (1000.),
           fISR       ( 1),
           fBeamStr   ( 1),
           fBeamWidth (0.002),
           fPole      (0.),
           fPolp      (0.),
	   fACC1      (0.05),
	   fACC2      (0.05),
	   fITMX1     (20),
	   fITMX2     (40),
           fPhotonPtr ( 0),
           fZBoost    (0.),
           fEcmIP     (fEcmInit),
           fEta       (0.),
           fPhi       (0.),
           fQ2AH      (0.),
           fJCombI    (0),
           fJCombF    (0),
           fSh        (0.),
           fCh        (0.),
           fR_BW_m    (0),
           fR_BW_p    (0),
           fR_BS_m    (0),
           fR_BS_p    (0),
           fR_ISR_var (0),
           fR_ISR_side(0)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init ahbases " << endl;

  using namespace std;
  stringstream ins(gJSF->Env()->GetValue("AHBases.Mass","125.")); 	 // M_h [GeV]
  ins >> fMass;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.Ecm","250."));       // E_cm [TeV]
  ins >> fEcmInit;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.Beamstrahlung","1")); // BmStr (on)
  ins >> fBeamStr;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.BeamWidth","0.002")); // Beam energy spread
  ins >> fBeamWidth;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.Bremsstrahlung","1"));// ISR (on)
  ins >> fISR;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.Pole","0."));         // electron polarization
  ins >> fPole;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.Polp","0."));         // positron polarization
  ins >> fPolp;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.ACC1","0.05"));
  ins >> fACC1;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.ACC2","0.01"));
  ins >> fACC2;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.ITMX1","40"));
  ins >> fITMX1;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.ITMX2","400"));
  ins >> fITMX2;

  ins.str("");
  ins.clear();
  ins.str(gJSF->Env()->GetValue("AHBases.NCALL","80000"));
  ins >> fNCALL;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Initial state helicity combination
  //--
  DefineVariable(fHelCombInitial, 0., 1., 0, 1);
  DefineVariable(fHelCombFinal  , 0., 1., 0, 1);

  //--
  //  xi, eta1, eata2, phi1, phi21:=phi2-phi1, q2ww
  //--
  DefineVariable(fXEta   , 0., 1., 1, 1);
  DefineVariable(fXPhi   , 0., 1., 1, 1);

  //--
  //  Beamstrahlung and natural Ebeam spread
  //--
  if (fBeamStr == 1) {
    DefineVariable(fR_BW_m, 0., 1., 0, 1);
    DefineVariable(fR_BW_p, 0., 1., 0, 1);
    DefineVariable(fR_BS_m, 0., 1., 0, 1);
    DefineVariable(fR_BS_p, 0., 1., 0, 1);
  }
 
  //--
  //  ISR
  //--
  if (fISR==1) {
    DefineVariable(fR_ISR_var,  0., 1., 0, 1);
    DefineVariable(fR_ISR_side, 0., 1., 0, 1);
  }

  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(fNCALL);

  SetTuneValue (1.5);
  SetIteration1(fACC1, fITMX1);
  SetIteration2(fACC2, fITMX2);
}
// --------------------------
//  D-tor
// --------------------------
AHBases::~AHBases()
{
  delete fPhotonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t AHBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.; // Jacobian factor
  Double_t eminus;        // E_e- after ISR and beamstrahlung
  Double_t eplus;         // E_e+ after ISR and beamstrahlung

  // --------------------------------------------
  //  Beamstrahlung
  // --------------------------------------------
  if (fBeamStr == 1) {
    bsWeight = fBM->GetWeight(fR_BW_m, fR_BW_p,
                              fR_BS_m, fR_BS_p,
                              eminus, eplus);
#if 0
    eplus   *= fBM->GetNominalEnergy();
    eminus  *= fBM->GetNominalEnergy();
#else
    eplus   *= fEcmInit/2.;
    eminus  *= fEcmInit/2.;
#endif
    fEcmIP   = 2.*TMath::Sqrt(eplus*eminus);     // ignore electron mass 
  } else {
    fEcmIP   = fEcmInit;      
    eplus    = fEcmInit/2.;
    eminus   = eplus;
  }
   
  // --------------------------------------------
  //  Initial State Radiation
  // --------------------------------------------

  static const Double_t kFisr = 2. * kAlpha0 / TMath::Pi();

  Double_t beta_isr = kFisr * (2.*(TMath::Log(fEcmIP/kM_e) - 1.));

  if (fISR == 1) {
    Double_t xisr    = TMath::Power(fR_ISR_var,1./beta_isr); // Ephoton / Ebeam 
             fEcmIP *= TMath::Sqrt(1. - xisr);                // reduced Ecm
    
    if (fR_ISR_side > 0.5) {
      eplus  *= 1. - xisr;
    } else {
      eminus *= 1. - xisr;
    }
  } // if Bremsstrahlung


  fZBoost = eminus - eplus; // P_z of the cm system after ISR and beamstrahlung

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

  Double_t m0 = 0.; // photon
  Double_t m1 = fMass;

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod.
  // --------------------------------------------
  if (fEcmIP < m0 + m1) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for e- is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;

  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------
  fQ2AH = fEcmIP*fEcmIP;

  // --------------------------------------------
  //  Handle photon emission.
  // --------------------------------------------
  Double_t delta = kM_e*kM_e/fQ2AH;
  Double_t etlo  = -TMath::Log(1+1/delta)/2.;
  Double_t ethi  =  TMath::Log(1+1/delta)/2.;
  fEta           = etlo + (ethi-etlo)*fXEta;
  bsWeight      *= ethi - etlo;

  Double_t exp2eta = TMath::Exp(2.*fEta);
           fSh     = TMath::Min(TMath::Sqrt((1.+2*delta)/(1.+exp2eta) - delta), 1.);
           fCh     = TMath::Min(TMath::Sqrt((1-fSh)*(1+fSh)), 1.);
           //fCh     = TMath::Min(TMath::Sqrt((exp2eta + delta*(exp2eta-1.))/(1.+exp2eta)),1.);

  Double_t costh = (fCh-fSh)*(fCh+fSh);
  weight         = (1.+2*delta)*4*exp2eta/(TMath::Power(exp2eta+1,2));
  bsWeight      *= weight;

  fPhi           = k2Pi*fXPhi;
  bsWeight      *= k2Pi;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  GENBranch cmbranch(fQ2AH, costh, fPhi, m0*m0, m1*m1);

  Double_t sigma = DSigmaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------

  Xh_fill( 1, fEcmIP           , (bsWeight*sigma));
  Xh_fill( 2, costh            , (bsWeight*sigma));
  Xh_fill( 3, fPhi             , (bsWeight*sigma));
  Xh_fill( 4, (Double_t)fJCombI, (bsWeight*sigma));
  Xh_fill( 5, (Double_t)fJCombF, (bsWeight*sigma));

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t AHBases::DSigmaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;

  Double_t q2a  = cmbranch.GetM12();
  Double_t q2h  = cmbranch.GetM22();
  Double_t cosa = cmbranch.GetCosTheta();
  Double_t phia = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2a, q2h, cmframe, cosa, phia, 0);
  fP[0] = phaseCM.GetFourMomentum(0); // photon
  fP[1] = phaseCM.GetFourMomentum(1); // Higgs

  Double_t betah = phaseCM.GetBetaBar();
  if (betah <= 0.) return 0.;

  Double_t eb     = rs/2.;
  Double_t pb     = TMath::Sqrt((eb-kM_e)*(eb+kM_e));
  Double_t beta_e = pb/eb;
  fK[0].SetXYZT(0., 0., pb, eb);
  fK[1].SetXYZT(0., 0.,-pb, eb);
#if 0
  cerr << "fP[0]=(" << fP[0](0) << ", " << fP[0](1) << ", " << fP[0](2) << ", " << fP[0](3) << ")" << endl; 
  cerr << "fP[1]=(" << fP[1](0) << ", " << fP[1](1) << ", " << fP[1](2) << ", " << fP[1](3) << ")" << endl; 
  ANL4DVector qcm = fP[0]+fP[1];
  cerr << "qcm=(" << qcm(0) << ", " << qcm(1) << ", " << qcm(2) << ", " << qcm(3) << ")" << endl; 
#endif
  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;

  // -------------------
  //  Amplitude squared
  // -------------------
  Double_t amp2 = AmpSquared(cmbranch);

  // -------------------
  //  Put them together
  // -------------------
  static const Int_t    kNbr  = 1;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));

  Double_t identp = 1.;                    // identical particle factor
  Double_t dPhase = kFact * betah;         // phase space factor
  Double_t flux   = 1./(2.* s * beta_e);   // beam flux factor
  Double_t spin   = 1.;                    // taken care of SelectHelicities

  Double_t sigma  = identp * flux * spin * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                              // now in [fb]

  return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
Double_t AHBases::AmpSquared(GENBranch &cmbranch)
{
  Complex_t amp   = FullAmplitude();
  Double_t  amp2  = TMath::Power(abs(amp),2);

  return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t AHBases::FullAmplitude()
{
   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion em(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELFermion ep(fK[1], kM_e, fHelInitial[1], -1, kIsOutgoing);

   HELVector  af(fP[0], 0., fHelFinal[0], +1); // photon
   HELScalar  hf(fP[1]); // Higgs

   Complex_t amp = AmpEEtoAH(em, ep, af, hf);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpEEtoAH()
// --------------------------
Complex_t AHBases::AmpEEtoAH(const HELFermion &em,
                             const HELFermion &ep,
                             const HELVector  &af,
                             const HELScalar  &hf)
{
   //-------------------
   // Coupling consts.
   //-------------------
   const Double_t  qe    = -1.;
   const Double_t  ge    = -qe*kGe;
   const Double_t  gal   =  ge;

   const Double_t  me    = kM_e;
   const Double_t  vev   = 2*kM_z/kGz;
   const Complex_t gye   = me/vev;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
#ifdef __PHASESPACE__
   //---------------------------
   // Just Phase Space
   //---------------------------
   Complex_t amp = 1;
#else
   //---------------------------
   // AH Production Amplitude
   //---------------------------
   //--
   // A from e-
   //--
   //   HELFermion(Double_t     eb,    // energy       of beam e+/-  [GeV]
   //              Double_t     ea,    // energy       of ISR photon [GeV]
   //	           Double_t     shlf,  // sin(theta/2) of ISR photon
   //	           Double_t     chlf,  // cos(theta/2) of ISR photon
   //	           Double_t     phi,   // phi          of ISR photon
   //	           Double_t     gal,   // QED coupling of e+/-
   //	           Double_t     me,    // electron mass
   //	           Int_t        nhe,   // helicity (-1,+1) of e+/-
   //              Int_t        nha,   // helicity (-1,+1) of ISR photon
   //	           Bool_t       isincom = kFALSE);
   //
   ANL4DVector pem  = em.GetFourMomentum();
   ANL4DVector pa   = af.GetFourMomentum();
   Double_t    eb   = pem(0);
   Double_t    ea   = pa (0);
   Double_t    phi  = fPhi;
   Double_t    chlf = fCh;
   Double_t    shlf = fSh;
   Int_t       nhe  = em.GetHelicity();
   Int_t       nhp  = ep.GetHelicity();
   Int_t       nha  = af.GetHelicity();

   //--
   // photon from e-
   //--
   static const Bool_t kIsIncoming = kTRUE;
   HELFermion ein(eb, ea, shlf, chlf, phi, nhe, nha, kIsIncoming, gal, me); 
   //HELFermion ein(em, af, gal, gal, me, 0.); 
   Complex_t  ampah = HELVertex(ein, ep, hf, gye, gye);

   //--
   // photon from e+
   //--
   static const Bool_t kIsOutgoing = kFALSE;
   HELFermion eot(eb, ea, shlf, chlf, phi, nhp, nha, kIsOutgoing, gal, me); 
   //HELFermion eot(ep, af, gal, gal, me, 0.); 
   Complex_t  ampha = HELVertex(em, eot, hf, gye, gye);

   //--
   // Sum of the two
   //--
   Complex_t amp = ampah + ampha;
#endif /* end __PHASESPACE__ */

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void AHBases::Userin()
{
  // --------------------------------------------
  //  Open beamstrahlung data
  // --------------------------------------------
  string bsfiledir;
  string bsfilename;

  if (fBeamStr) {
    stringstream ins(gJSF->Env()->GetValue("AHBases.BeamstrahlungFilepath",
                                  "/proj/soft/jsf/pro/data/bsdata/"));
    ins >> bsfiledir;

    ins.clear();
    ins.str(gJSF->Env()->GetValue("AHBases.BeamstrahlungFilename","trc500"));
    ins >> bsfilename;
    
    TFile *fBeamFile = new TFile((bsfiledir+bsfilename+".root").data());
    if (!fBeamFile) {
      cerr << " Unable to open a file for beamstrahlung" << endl;
      abort();
    }
    
  // --------------------------------------------
  //  Initialize beam generator
  // --------------------------------------------

    fBM = (JSFBeamGenerationCain*)fBeamFile->Get(bsfilename.data());
    fBM->SetIBParameters(fBeamWidth);

    fBM->MakeBSMap();
    fBM->Print(); 

    cout << " Nominal Energy (beamstrahlung) = " << fBM->GetNominalEnergy() 
         << " beam width     (beamstrahlung) ="  << fBM->GetIBWidth() 
         << endl;

    if (!(fBM->GetNominalEnergy() == (fEcmInit/2))) {
      cout << "Nominal energy from beamstrahung is " << fBM->GetNominalEnergy() 
           << " which is different from fEcm/2: "    << (fEcmInit/2)
           << endl;        
    } // check the energy homogeneity
  } // if beamstrahlung is on


  if (!fPhotonPtr) fPhotonPtr = new GENPDTPhoton();
  //fPhotonPtr->DebugPrint();

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  Xh_init( 1,     0., fEcmInit*1.1, 50, "Ecm"   );
  Xh_init( 2,    -1.,    +1.,       50, "Costh" );
  Xh_init( 3,     0.,   k2Pi,       50, "Phi"   );
  Xh_init( 4,     0.,     4.,        4, "Helin ");
  Xh_init( 5,     0.,     2.,        2, "Helot ");
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void AHBases::Userout()
{
  cout << "End of AHBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcmInit << " [GeV]   "    << endl
       << "Beamstrahlung        = " << (fBeamStr ? "on" : "off")  << endl
       << "Bremsstrahlung       = " << (fISR     ? "on" : "off")  << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"  << endl
       << "Number of iterations = " << GetNoOfIterate()           << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void AHBases::SelectHelicities(Double_t &weight)
{
   static const Int_t kNi = 4;
   static const Int_t kIHelComb[kNi][2] = {{-1, -1},
                                           {-1, +1},
                                           {+1, -1},
                                           {+1, +1}};
   static const Int_t kNf = 2;
   static const Int_t kFHelComb[kNf][2] = {{-1,  0},
                                           {+1,  0}};
   Double_t helm = (1. - fPole)/2.;
   Double_t help = (1. - fPolp)/2.;
   if (fHelCombInitial < helm || helm == 1.) {
      Double_t helcombi = fHelCombInitial/helm;
      if (helcombi < help) fJCombI = 0;
      else                 fJCombI = 1;
   } else {
      fJCombI = 1;
      Double_t helcombi = (fHelCombInitial-helm)/(1.-helm);
      if (helcombi < help) fJCombI = 2;
      else                 fJCombI = 3;
   }
   fHelInitial[0] = kIHelComb[fJCombI][0];
   fHelInitial[1] = kIHelComb[fJCombI][1];
   fJCombF = static_cast<Int_t>(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   weight = kNf;
}
