#ifndef EASPRING_H
#define EASPRING_H
//*****************************************************************************
//* =====================
//*  EASpring
//* =====================
//*  
//* (Description)
//*    e- gamma -> e- gamma generator
//*
//* (Update Record)
//*    2019/12/25  K.Fujii	Original version.
//*
//*****************************************************************************

#include "TNamed.h"
#include "JSFModule.h"
#include "JSFBases.h"
#include "JSFSpring.h"
#include "JSFBeamGeneration.h"

#include "HELLib.h"
#include "GENLib.h"

//_______________________________________________________________________
// =====================
//  class EABases
// =====================
//-----------------------------------------------------------------------
class EABases : public JSFBases {
friend class EASpringBuf;
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  EABases(const char *name  = "EABases", 
          const char *title = "EA Bases");
  virtual ~EABases();

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetEeminus  ()           const { return fEminus;   }
  Double_t GetPeminus  ()           const { return fPminus;   }
  Double_t GetEgamma   ()           const { return fEgamma;   }

  Double_t GetCosTheta ()           const { return fCosTheta;  }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  void     SetPeminus  (Double_t pem    ) { fPminus    = pem;  }
  void     SetEgamma   (Double_t egm    ) { fEgamma    = egm;  }
  void     SetPole     (Double_t p      ) { fPole      = p;    }
  void     SetPola     (Double_t p      ) { fPola      = p;    }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual void     Userin();   // Bases user initialization
  virtual void     Userout();  // Bases user output 

  Double_t Func();     // Bases integration function.

  // ----------------------
  //   Utility methods
  // ----------------------
private:
  void     SelectHelicities(Double_t &weight);

  Double_t  DSigmaDX     ();
  Double_t  AmpSquared   ();
  Complex_t FullAmplitude();
  Complex_t AmpEAtoEA    (const HELFermion &em,
                          const HELVector  &vi,
                          const HELFermion &f,
                          const HELVector  &vo);

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Job parameters
  // ----------------
  Double_t fPminus;         // Initial p_e-
  Double_t fEminus;         // Initial E_e-
  Double_t fEgamma;         // Initial E_gamma
  Double_t fPole;           // electron polarization
  Double_t fPola;           // photon   polarization

  // ----------------
  //  Particle Data
  // ----------------
  GENPDTEntry  *fFPtr;            //! PD table entry of "f"

  // ----------------
  //  Event info
  // ----------------
  Double_t       fZBoost;         // p_z(cm)      in lab frame
  Double_t       fEcmIP;          // Ecm after B-strahlung
  Int_t          fHelInitial[2];  // initial state helicities
  Int_t          fHelFinal  [2];  // final   state helicities
  Int_t          fJCombI;         // intial helicity combination
  Int_t          fJCombF;         // final  helicity combination
  ANL4DVector    fK[2];           // [0,1] = [e-, a]
  ANL4DVector    fP[2];           // [0,1] = [e-, a]

  // ----------------
  //  Integ. vars
  // ----------------
  Double_t       fHelCombInitial; // initial state helicity combination
  Double_t       fHelCombFinal;   // final   state helicity combination
  Double_t       fCosTheta;       // cos(theta_x) in cm  frame
  Double_t       fPhi;            // phi_x        in cm  frame

  Double_t       fXU[2];          //! [0,1] = (cosx, phix)
  Double_t       fXL[2];          //!

  ClassDef(EABases, 1) // Bases for e+e- -> XX process
};

class EASpring;

//_______________________________________________________________________
// =====================
//  class EASpringBuf
// =====================
//-----------------------------------------------------------------------
class EASpringBuf : public JSFSpringBuf {
public:
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
  EASpringBuf(const char *name   = "EASpringBuf", 
              const char *title  = "EA Spring test event buffer",
                EASpring *spring = 0)
            : JSFSpringBuf(name,title,(JSFSpring*)spring) {} 
  virtual ~EASpringBuf() {}

  // ----------------------
  //  Getters and Setters
  // ----------------------
  Double_t GetZBoost   ()           const { return fZBoost;    }
  Double_t GetCosTheta ()           const { return fCosTheta;  }
  Double_t GetPhi      ()           const { return fPhi;       }
  Double_t GetEcmIP    ()           const { return fEcmIP;     }

  // ----------------------
  //   Base class methods
  // ----------------------
  virtual Bool_t SetPartons();

private:
  // --------------------------------------------------------------------
  //  Data Members
  // --------------------------------------------------------------------
  // ----------------
  //  Event info
  // ----------------
  Double_t fZBoost;         // p_z(cm)      in lab frame
  Double_t fCosTheta;       // cos(theta_x) in cm  frame
  Double_t fPhi;            // phi_x        in cm  frame
  Double_t fEcmIP;          // Ecm after B-strahlung

  ClassDef(EASpringBuf, 1)  // EASpring event buffer
};

//_______________________________________________________________________
// =====================
//  class EASpring
// =====================
//-----------------------------------------------------------------------
class EASpring : public JSFSpring {
  // --------------------------------------------------------------------
  //  Member Functions
  // --------------------------------------------------------------------
  // ----------------------
  //  C-tor and D-tor
  // ----------------------
public:
   EASpring(const char *name  = "EASpring", 
              const char *title = "EA Spring test",
              EABases  *bases = 0);
   virtual ~EASpring();

   ClassDef(EASpring, 1)  // EASpring class
};
#endif
