#include <sstream>
#include <iomanip>
#include "TGraph.h"
#include "ANL4DVector.h"
//#define __DEBUG__

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"
#include "GENLib.h"
#include "HELLib.h"
using namespace std;

ANL4DVector fK[2];
ANL4DVector fP[2];
int         fHelInitial[2];  // initial state helicities
int         fHelFinal  [2];  // final   state helicities

//_____________________________________________________________________________
// --------------------------
//  AmpEAtoEA()
// --------------------------
Complex_t AmpEAtoEA(const HELFermion &ei,
                    const HELVector  &vi,
		    const HELFermion &eo,
		    const HELVector  &vo)
{
   double  qe    = -1.;
   double  ge    = qe*kGe;

   //--------------------------------------------------
   // Calculate Amplitudes
   //--------------------------------------------------
   HELFermion eai(ei, vi, ge, ge, kM_e, 0.);
   Complex_t ampa = HELVertex(eai, eo, vo, ge, ge);

   HELFermion eao(ei, vo, ge, ge, kM_e, 0.);
   Complex_t ampat = HELVertex(eao, eo, vi, ge, ge);

   Complex_t amp = ampa + ampat;

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  FullAmplitude()
// --------------------------
Complex_t FullAmplitude()
{
   static const Bool_t kIsIncoming = kTRUE;
   static const Bool_t kIsOutgoing = kFALSE;

   HELFermion ei(fK[0], kM_e, fHelInitial[0], +1, kIsIncoming);
   HELVector  vi(fK[1],   0., fHelInitial[1], -1);

   HELFermion eo(fP[0], kM_e, fHelFinal  [0], +1, kIsOutgoing);
   HELVector  vo(fP[1],   0., fHelFinal  [1], +1);

   Complex_t amp = AmpEAtoEA(ei, vi, eo, vo);

   return amp;
}

//_____________________________________________________________________________
// --------------------------
//  AmpSquared()
// --------------------------
double AmpSquared()
{
   Complex_t amp   = FullAmplitude();
   double    amp2  = TMath::Power(abs(amp),2);
 
   return amp2;
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
double DSigmaDX(double rs, double cs, double pole, double pola)
{
   // --------------------------------------------
   //  Phase space
   // --------------------------------------------
   // fEcmIP, fCosTheta, fPhi, fFPtr);
   double cosx = cs;
   double phix = 0.;
   double mf   = kM_e;
   double mf2  = mf*mf;
 
   ANL4DVector qcm(rs, 0., 0., 0.);
   GENFrame    cmframe;
 
   GENPhase2 phaseCM(qcm, mf2, 0., cmframe, cosx, phix, 0);
   double betax = phaseCM.GetBetaBar();
   if (betax <= 0.) return 0.;
   fP[0] = phaseCM.GetFourMomentum(0);
   fP[1] = phaseCM.GetFourMomentum(1);
 
   double betab  = (1.-kM_e/rs)*(1.+kM_e/rs);
   double pb     = (rs/2)*betab;
   double eb     = TMath::Sqrt(pb*pb+kM_e*kM_e);
   double ea     = pb;
   double beta_e = pb/eb;
   fK[0].SetXYZT(0., 0., pb, eb);
   fK[1].SetXYZT(0., 0.,-pb, ea);
 
   // --------------------------------------------
   //  Calcuate differential cross section
   // --------------------------------------------
#ifdef __DEBUG__
   cerr << " fP[0] = (" 
        << fP[0].E () << ","
        << fP[0].Px() << ","
        << fP[0].Py() << ","
        << fP[0].Pz() << ")" << endl;
   cerr << " fP[1] = (" 
        << fP[1].E () << ","
        << fP[1].Px() << ","
        << fP[1].Py() << ","
        << fP[1].Pz() << ")" << endl;
   ANL4DVector pcm = fP[0] + fP[1];
   cerr << " pcm = (" 
        << pcm.E () << ","
        << pcm.Px() << ","
        << pcm.Py() << ","
        << pcm.Pz() << ")" << endl;
#endif

   // -------------------
   //  Amplitude squared
   // -------------------
   double amp2 = 0;
   for (int helei=0; helei<2; helei++) {
      fHelInitial[0] = -1+2*helei;
      double wat = helei == 0 ? (1.-pole)/2. : (1.+pole)/2.;
      for (int helai=0; helai<2; helai++) {
         fHelInitial[1] = -1+2*helai;
         double wat  = helei == 0 ? (1.-pole)/2. : (1.+pole)/2.;
                wat *= helai == 0 ? (1.-pola)/2. : (1.+pola)/2.;
         for (int heleo=0; heleo<2; heleo++) {
            fHelFinal[0] = -1+2*heleo;
            for (int helao=0; helao<2; helao++) {
               fHelFinal[1] = -1+2*helao;
               double amp2hel = AmpSquared();
 	              amp2   += amp2hel*wat;
 	    }
         }
      }
   }
 
   // -------------------
   //  Put them together
   // -------------------
   static const Int_t    kNbr  = 1;
   static const double kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));
 
   double identp = 1.;                            // identical particle factor
   double dPhase = kFact * betax;                 // phase space factor
   double flux   = 1./(4.* eb * ea * (1.+beta_e));// beam flux factor
   double spin   = 1.;                            // spin average for photon
 
   double sigma  = identp * flux * spin * amp2 * dPhase; // in [1/GeV^2]
          sigma *= kGeV2fb;                              // now in [fb]
 
   return sigma;
}

//_____________________________________________________________________________
// --------------------------
//  Function: eaplot
// --------------------------
TGraph *eaplot(double pminus = 20.000, // p_e- in lab frame
               double egamma =  3.e-9, // E_gamma in lab frame
               double pole   =  0.000, // electron polarization
               double pola   =  0.000, // photon polarizationn
 	       double csmin  = -0.999, // Min(cos(theta_e)) in CM frame
	       double csmax  = +0.999) // Max(cos(theta_e)) in CM frame
{
   // --------------------------------------------
   //  Select unit for cross section.
   // --------------------------------------------
   const double kfb  = 1.;
   const double kpb  = 1.e-3;
   const double kmb  = 1.e-12;
   const double kb   = 1.e-15;
         double unit = kmb;

   // --------------------------------------------
   //  Calculate Ecm, etc.
   // --------------------------------------------
   double mf      = kM_e;
   double boost   = pminus - egamma; // P_z of the cm system
   double eminus  = TMath::Sqrt(pminus*pminus+kM_e*kM_e);
   double rs      = TMath::Sqrt((eminus+egamma-boost)*(eminus+pminus));

   double pstar   = (rs/2)*(1-kM_e/rs)*(1+kM_e/rs);
   double emstar  = TMath::Sqrt(kM_e*kM_e+pstar*pstar);
   double eastar  = pstar;
   double bestar  = pstar/emstar;
   double be      = pminus/eminus;
   double fact    = emstar*eastar*(1.+bestar)/(eminus*egamma*(1.+be)); // from CM to Lab
   double gam     = (eminus+egamma)/rs;      // gamma of CM system
   double bet     = (boost)/(eminus+egamma); // beta of CM system
   double jacob   = 1./(gam*bet*pstar);     // Jacobian from dcos* to dealab
          jacob  *= pminus > egamma ? 1. : -1.;
 
   // --------------------------------------------
   //  Check if reduced Ecm enough for prod.
   // --------------------------------------------
   if (rs < mf) {
     return 0;
   }
 
   // --------------------------------------------
   //  Calcuate differential cross section
   // --------------------------------------------
   static const int kNpts = 101;
   double xdata[kNpts], ydata[kNpts];
   int    npts   = kNpts - 1;
   double dcs    = (csmax-csmin)/npts;
   double sigma  = 0.;
   double wt     = 1.;
   for (int i=0; i<=npts; i++) {
      if (i == 0 || i == npts)   wt = 1./3;
      else if(i%2 == 0)          wt = 2./3;
      else                       wt = 4./3;
      double cs      = csmin + i*dcs;
      double dsigdx = DSigmaDX(rs,cs,pole,pola);
      sigma += dsigdx * wt;
#if 0
      cerr << " cs = " << cs << " dsig = " << dsigdx * unit << endl;
      xdata[i] = cs;
      ydata[i] = dsigdx * unit;
#else
      double ea = gam*pstar*(1-bet*cs);		    // ealab
      double dsigdea = dsigdx * fact * jacob * k2Pi; // to dsigma/dea_lab
      cerr << " ea = " << ea << " dsig = " << dsigdea * unit << endl;
      xdata[i] = ea;
      ydata[i] = dsigdea * unit;
#endif
   }
   sigma *= k2Pi*dcs; // phi integral
   TGraph *grp = new TGraph(kNpts,xdata,ydata);
 
   // --------------------------------------------
   //  Print out results.
   // --------------------------------------------
   sigma *= unit;
   cout << "End of eaplot ----------------------------------- "         << endl
        << "Peminus              = "  << pminus  << " [GeV]   "         << endl
        << "Eegamma              = "  << egamma  << " [GeV]   "         << endl
        << "Ecm                  = "  << rs      << " [GeV]   "         << endl
        << "(Pole,Pola)          = (" << pole    << ", " << pola << ")" << endl
        << "Total Cross section  = "  << sigma   << " [mb]    "         << endl;
 
   // --------------------------------------------
   //  That's it.
   // --------------------------------------------
   return grp;
 }
