//*************************************************************************
//* -----------
//* MyPFOAnalysis
//* -----------
//* (Description)
//* 	Classes to analyze PFO data.
//* (Update Record)
//*	2006/12/20  K.Fujii	Original version.
//*************************************************************************

#include "JSFSteer.h"
#include "MyPFOAnalysis.h"
#include "MyPrimaryParticle.h"
#include <iostream>
#include <sstream>

#include "S4PFOMakerBuf.h"
#include "S4GENMCTruthBuf.h"
#include "S4PrimaryVertex.h"
#include "S4JetMakerBuf.h"
#include "S4JetSet.h"
#include "S4Jet.h"

#include "TNtupleD.h"
#if __GNUC__ >= 5 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3)
#include <cstdlib>
#endif

//#define __BPLOCK__
//#define __NOBPLOCK__
#define __CORRECTMP__

ClassImp(MyPFOAnalysis)

//___________________________________________________________________________
MyPFOAnalysis::MyPFOAnalysis(const Char_t *name, const Char_t *title)
             : JSFModule(name, title)
{
   //std::cerr << " constructor for MyPFOAnalysis is called..." << std::endl;
}

//___________________________________________________________________________
MyPFOAnalysis::~MyPFOAnalysis()
{
}

//___________________________________________________________________________
Bool_t MyPFOAnalysis::Initialize()
{
#if defined(__NOBPLOCK__)
   cerr << " NOBPLOCK = YES ----------------------------- " << endl;
#endif
#if defined(__BPLOCK__) && !defined(__NOBPLOCK__)
   cerr << " BPLOCK = YES && NOBPLOCK = NO -------------- " << endl;
#endif
#if !defined(__BPLOCK__) && !defined(__NOBPLOCK__)
   cerr << " BPLOCK = NO && NOBPLOCK = NO -------------- " << endl;
#endif

   return kTRUE;
}


//___________________________________________________________________________
Bool_t MyPFOAnalysis::Process(Int_t ev)
{
   fPrimaryBPs.Clear();		// remove old previous primary BPs
   MonitorPFOs(ev);		// PFO analysis
#if 1
   std::cerr << "MyPFOAnalysis::Process evt " << ev << " is finished" << std::endl;
#endif
   return kTRUE;   
}   

//___________________________________________________________________________
void MyPFOAnalysis::MonitorPFOs(Int_t ev)
{
   static Int_t nev = 0;
   nev++;

   TDirectory *last = gDirectory;
   gFile->cd("/");

   TObjArray &pbps = GetPrimaryBPs(); 	// do this here to flag primary BPs beforehand

   TObjArray *outputbufp = static_cast<S4PFOMakerBuf *> 
     (gJSF->FindModule("S4PFOMaker")->EventBuf())->GetOutputBufPtr();

   Int_t Npfo = outputbufp->GetEntries();
   if (!Npfo) return; 			// quit if no PFOs

   // -------------------------
   //  Lock all decendants
   // -------------------------
   // Too idealistic treatment to avoid double counting of daughter energy when paraent 
   // is already taken into account: this means that any neutral kink daughter can be
   // identified as a kink daughter, for instance.
   TIter nextpfo(outputbufp);
   S4PFObject *pfop;
#if defined(__BPLOCK__) && !defined(__NOBPLOCK__)
   while ((pfop = static_cast<S4PFObject *> (nextpfo()))) {
#if 0
     S4HYBTrack *tkp = pfop->GetHYBTrackPtr();
     if (pfop->IsLocked() || (tkp && tkp->TestBit(BIT(16)))) { // BIT(16) = kink mother
       S4BreakPoint *bpp  = pfop->GetBreakPointPtr();
       if (bpp) bpp->LockAllDescendants();
     }
#else
     S4BreakPoint *bpp  = pfop->GetBreakPointPtr();
     if (bpp) bpp->LockAllDescendants();
#endif
   }
#endif

   // -------------------------
   //  Get jet information
   // -------------------------
   const Int_t kNjets = 2;		// select 2-jet event
   S4JetSet   *jsp  = GetJetSet(kNjets);
   Int_t       npj [kNjets];		// # particles in jet
   Double_t    cosj[kNjets];		// jet angle
   ANL4DVector qvis;			// visible 4-momentum
   if (jsp) {
     for (Int_t i=0; i<kNjets; i++) {
       if (jsp) {
         S4Jet *jp = static_cast<S4Jet *>(jsp->At(i));
         npj [i] = jp->GetNparticles();
         cosj[i] = jp->CosTheta();
	 qvis += *jp;
       } else {
         npj [i] = 0;
         cosj[i] = -99999.;
       }
     }
   }
   Double_t evis = qvis.E();
   Double_t pt   = qvis.Pt();
   Double_t pl   = qvis.Z();
   Double_t mass = qvis.Mag();

   static const UInt_t kUsedInJets = BIT(22);	// used in jet finder
   nextpfo.Reset();
   while ((pfop = static_cast<S4PFObject *> (nextpfo()))) {
     if (!pfop->IsLocked()) pfop->SetBit(kUsedInJets);
   }

#if !defined(__BPLOCK__) && !defined(__NOBPLOCK__)
   nextpfo.Reset();
   while ((pfop = static_cast<S4PFObject *> (nextpfo()))) {
#if 0
     S4HYBTrack *tkp = pfop->GetHYBTrackPtr();
     if (pfop->IsLocked() || (tkp && tkp->TestBit(BIT(16)))) { // BIT(16) = kink mother
       S4BreakPoint *bpp  = pfop->GetBreakPointPtr();
       if (bpp) bpp->LockAllDescendants();
     }
#else
     S4BreakPoint *bpp  = pfop->GetBreakPointPtr();
     if (bpp) bpp->LockAllDescendants();
#endif
   }
#endif

   // -------------------------
   //  Unlock primary BPs
   // -------------------------
   TIter nextpbp(&pbps);
   S4BreakPoint *pbpp;
   while ((pbpp = static_cast<S4BreakPoint *>(nextpbp()))) pbpp->Unlock();

   // -------------------------
   //  Create Ntuples
   // -------------------------
   static TNtupleD *tupp   = 0; 		// pfo by pfo information
   if (!tupp) tupp = new TNtupleD("hpfo", "", "ev:npj1:npj2:cosj1:cosj2:evis:pt:pl:mass:ch:pid:epfo:ecl:etk:epos:ebp:epbp:pxpfo:pypfo:pzpfo:gppid:ppid:vx:vy:vz");

   static TNtupleD *tupdp  = 0; 		// double counted pfo by pfo information
   if (!tupdp) tupdp = new TNtupleD("hdbl", "", "ev:npj1:npj2:cosj1:cosj2:evis:pt:pl:mass:ch:pid:epfo:ebp:pxpfo:pypfo:pzpfo:vxbp:vybp:vzbp");

   static TNtupleD *tupahp = 0; 		// undetected particle information
   if (!tupahp) tupahp = new TNtupleD("hah", "", "ev:npj1:npj2:cosj1:cosj2:evis:pt:pl:mass:ch:pid:ppid:e:px:py:pz");

   static TNtupleD *tupv0p = 0; 		// v0 by v0 information
   if (!tupv0p) tupv0p = new TNtupleD("hv0", "", "ev:npj1:npj2:cosj1:cosj2:evis:pt:pl:mass:pid:qual:vx:vy:vz:px:py:pz:e:vxbp:vybp:vzbp:pxbp:pybp:pzbp:ebp");

   static TNtupleD *tupep  = 0; 		// event information
   if (!tupep) tupep = new TNtupleD("hevt", "", "ev:npj1:npj2:cosj1:cosj2:evis:pt:pl:mass:npfos:eem:ehd:etk:eembp:ehdbp:etkbp:eempbp:ehdpbp:etkpbp:eah:eahznu:eahisr:eisr:eisrbp:eisrpbp:px:py:pz:pxisr:pyisr:pzisr:pxah:pyah:pzah:pxahznu:pyahznu:pzahznu:pxahisr:pyahisr:pzahisr:eahhnu:pxahhnu:pyahhnu:pzahhnu:pxem:pyem:pzem:pxhd:pyhd:pzhd:edbltk:pxdbltk:pydbltk:pzdbltk:edblem:pxdblem:pydblem:pzdblem:edblhd:pxdblhd:pydblhd:pzdblhd:edbltkab:pxdbltkab:pydbltkab:pzdbltkab:etkmpc:pxtkmpc:pytkmpc:pztkmpc:edbltkmpc:pxdbltkmpc:pydbltkmpc:pzdbltkmpc:etkbpmpc");

   // -------------------------
   //  Detected particles
   // -------------------------
   Int_t    Npfos  = 0;		// # unlocked PFOs with breakpoint
   ANL4DVector qtot;		// Ptot   = sigma_{i=1}^{npfos} ppfo_i
   ANL4DVector qem;		// Pem    = sigma_{i in EM && !htkp} ppfo_i
   ANL4DVector qhd;		// Pem    = sigma_{i in HD && !htkp} ppfo_i
   ANL4DVector qtk;		// Ptk    = sigma_{i in TK}          ppfo_i
   ANL4DVector qisr;		// Pisr   = sigma_{i from isr}       ppfo_i
   ANL4DVector qembp;		// Pem    = sigma_{i in EM && !htkp} pbp_i
   ANL4DVector qhdbp;		// Phd    = sigma_{i in HD && !htkp} pbp_i
   ANL4DVector qtkbp;		// Ptk    = sigma_{i in TK}          pbp_i
   ANL4DVector qisrbp;		// Pisbp  = sigma_{i from isr}       pbp_i
   ANL4DVector qempbp;		// Pem    = sigma_{i in EM && !htkp} ppbp_i
   ANL4DVector qhdpbp;		// Phd    = sigma_{i in HD && !htkp} ppbp_i
   ANL4DVector qtkpbp;		// Ptk    = sigma_{i in TK}          ppbp_i
   ANL4DVector qisrpbp;		// Pisbp  = sigma_{i from isr}       ppbp_i
   ANL4DVector qtkmpc;		// Ptk    = sigma_{i in TK}          ppfo_i with Mp correction
   ANL4DVector qtkbpmpc;	// Ptk    = sigma_{i in TK}          pbp_i  with Mp correction

   ANL4DVector qdbltk;		// double counted TK pfos
   ANL4DVector qdblem;		// double counted EM pfos
   ANL4DVector qdblhd;		// double counted HD pfos
   ANL4DVector qdbltkab;	// double counted albedo TK pfos
   ANL4DVector qdbltkmpc;	// double counted TK pfos with Mp correction

   nextpfo.Reset();
   while ((pfop = static_cast<S4PFObject *> (nextpfo()))) {
     S4CALCluster *clp    = pfop->GetClusterPtr();
     S4HYBTrack   *htkp   = pfop->GetHYBTrackPtr();
     Int_t         pidpfo = pfop->GetPDGCode();
     S4BreakPoint *bpp    = pfop->GetBreakPointPtr();
     S4BreakPoint *pbpp   = pfop->GetPrimaryParticleBreakPoint();
     if (pfop->IsLocked() || !bpp) {		// skip locked or particles with BP
       if (pfop->TestBit(kUsedInJets)) { 	// seledt double countd PFOs
         if (htkp || (!htkp && !clp)) { 			// charged PFO or V0
           qdbltk += *pfop;
	   if (!bpp) qdbltkab += *pfop;				// albedo tk PFO
#ifdef __CORRECTMP__
           ANL4DVector qpfompc = *pfop;
	   if (bpp && pidpfo == 2212 /* p */ && bpp != pbpp) { // 2ndary protons
             qpfompc.SetE(qpfompc.E()-qpfompc.Mag());
	   }
	   qdbltkmpc += qpfompc;
#endif
         } else {           					// neutral PFO
           if (pidpfo == 22 || TMath::Abs(pidpfo) == 11) {	// EM CAL
             qdblem += *pfop;
           } else {						// hadron CAL
             qdblhd += *pfop;
	   }
	 }
       }
       Double_t ebp = bpp ? bpp->GetFourMomentum().E() : 0.;
       if ((!bpp || bpp != pbpp) && (pidpfo == 2112 /* n */)) {
         static const Double_t kMn = 0.93956536;
         ebp -= (bpp ? bpp->GetFourMomentum().Mag() : kMn);     // don't count mass for partciles kicked off materials
       }
       Double_t chpfo  = pfop->GetCharge();
       TVector3 vtx    = bpp ? bpp->GetPosition() : TVector3(0.,0.,0.);
       static Int_t kNdbdat = 19;
       Double_t dbdata[kNdbdat];
       dbdata[ 0] = nev;	// event number
       dbdata[ 1] = npj[0];	// # pfos     in jet1
       dbdata[ 2] = npj[1];	// # pfos     in jet2
       dbdata[ 3] = cosj[0];	// cos(theta) of jet1
       dbdata[ 4] = cosj[1];	// cos(theta) of jet2
       dbdata[ 5] = evis;	// Evis       of the final state
       dbdata[ 6] = pt;		// Pt         of the final state
       dbdata[ 7] = pl;		// Pl         of the final state
       dbdata[ 8] = mass;	// inv. mass  of the final state
       dbdata[ 9] = chpfo;	// charge             of this pfo
       dbdata[10] = pidpfo;	// pid                of this pfo
       dbdata[11] = pfop->E();	// energy             of this pfo
       dbdata[12] = ebp;	// break point energy of this pfo
       dbdata[13] = pfop->Px();	// px                 of this pfo
       dbdata[14] = pfop->Py();	// py                 of this pfo
       dbdata[15] = pfop->Pz();	// pz                 of this pfo
       dbdata[16] = vtx.X();	// vx                 of this pfo
       dbdata[17] = vtx.Y();	// vy                 of this pfo
       dbdata[18] = vtx.Z();	// vz                 of this pfo
       tupdp ->Fill(dbdata);
       continue;
     }
     Npfos++;
     Double_t      chpfo  = pfop->GetCharge();
     S4VExactHit  *posp   = pfop->GetExactHitPtr();
     Double_t      epos   = posp ? posp->GetTotalEnergy() : 0.;
     Double_t      ecl    = clp ? clp->GetEnergy  ()     : 0.;
     Double_t      etk    = htkp||(!htkp&&!clp) ? pfop->E () : 0.;
     ANL4DVector   qbp    = bpp->GetFourMomentum();
     Double_t      ebp    = qbp.E();
#if 0
#if 0
     if (bpp != pbpp && (pidpfo == 11 /* e */ || pidpfo == 2212 /* p */ || pidpfo == 2112 /* n */ || pidpfo == 0 /* nuclei */)) {
#else
     // When bpp is a 2ndary particle, assume it is kicked off materials if it is an e- or a nucleus,
     // or if it is a proton or a neutron from a parent which is not a nucleon.
#if 0
     Int_t         pidpbp = pbpp ? pbpp->GetParticleID() : 0;
     if (bpp != pbpp && ((pidpfo == 11 /* e */ || pidpfo == 0 /* nuclei */) || 
	                ((pidpbp != 2212 && pidpbp != 2112) && (pidpfo == 2212 /* p */ || pidpfo == 2112 /* n */)))) {
#else
     Int_t j4tkidipbp = bpp->GetParentID() > pbpp->GetTrackID() ? bpp->GetParentID() : pbpp->GetTrackID();
     S4BreakPoint *ipbpp   = S4BreakPoint::GetBreakPointPtr(j4tkidipbp);
     Int_t         pidipbp = ipbpp->GetParticleID();
     if (bpp != ipbpp && ((pidpfo == 11 /* e */ || pidpfo == 0 /* nuclei */) || 
	                ((pidipbp != 2212 && pidipbp != 2112) && (pidpfo == 2212 /* p */ || pidpfo == 2112 /* n */)))) {
#endif
#endif
#else
     if (bpp != pbpp && (pidpfo == 2112 /* n */)) {
       ebp -= qbp.Mag();     // don't count mass for partciles kicked off materials
       qbp.SetE(ebp);
#endif
     }
     Double_t epbp   = pbpp->GetFourMomentum().E();
     qtot += *pfop;
     if (htkp || (!htkp && !clp)) { 	// charged PFO or V0
       qtk   += *pfop;
       qtkbp += qbp;
       if (!pbpp->IsLocked()) qtkpbp += pbpp->GetFourMomentum();
#ifdef __CORRECTMP__
       ANL4DVector qpfompc = *pfop;
       ANL4DVector qbpmpc  = qbp;
       if (pidpfo == 2212 /* p */ && bpp != pbpp) { // 2ndary protons
         qpfompc.SetE(qpfompc.E()-qpfompc.Mag());
         qbpmpc.SetE(qbpmpc.E()-qbpmpc.Mag());
       }
       qtkmpc   += qpfompc;
       qtkbpmpc += qbpmpc;
#endif
     } else {           		// neutral PFO
       if (pidpfo == 22 || TMath::Abs(pidpfo) == 11) {// EM CAL
         qem   += *pfop;
         qembp += qbp;
         if (!pbpp->IsLocked()) qempbp += pbpp->GetFourMomentum();
       } else {				// hadron CAL
         qhd   += *pfop;
         qhdbp += qbp;
         if (!pbpp->IsLocked()) qhdpbp += pbpp->GetFourMomentum();
       }
     }

     Int_t j4tkidpbp = pbpp->GetTrackID();
     MyPrimaryParticle *myppp  = MyPrimaryParticle::GetMyPPPtr(j4tkidpbp);
     MyPrimaryParticle *mygppp = myppp  
                               ? (myppp->GetGrandParentPtr() ? myppp->GetGrandParentPtr() : myppp)
                               : 0;
     Double_t           gppid  = mygppp ? mygppp->GetParticleID() : 0;
     Int_t              pidpbp = pbpp->GetParticleID();
     if (TMath::Abs(gppid-22.)<0.1) { // isr photons
       qisr   += *pfop;
       qisrbp += qbp;
       if (!pbpp->IsLocked()) qisrpbp += pbpp->GetFourMomentum();
     }
     TVector3 vtx    = bpp->GetPosition();

     static Int_t kNpfdat = 25;
     Double_t pfdata[kNpfdat];
     pfdata[ 0] = nev;		// event number
     pfdata[ 1] = npj[0];	// # pfos     in jet1
     pfdata[ 2] = npj[1];	// # pfos     in jet2
     pfdata[ 3] = cosj[0];	// cos(theta) of jet1
     pfdata[ 4] = cosj[1];	// cos(theta) of jet2
     pfdata[ 5] = evis;		// Evis       of the final state
     pfdata[ 6] = pt;		// Pt         of the final state
     pfdata[ 7] = pl;		// Pl         of the final state
     pfdata[ 8] = mass;		// inv. mass  of the final state
     pfdata[ 9] = chpfo;	// charge             of this pfo
     pfdata[10] = pidpfo;	// pid                of this pfo
     pfdata[11] = pfop->E();	// energy             of this pfo
     pfdata[12] = ecl;		// cluster     energy of this pfo
     pfdata[13] = etk;		// track       energy of this pfo
     pfdata[14] = epos;		// post hit    energy of this pfo
     pfdata[15] = ebp;		// break point energy of this pfo
     pfdata[16] = epbp;		// primary bp  energy of this pfo
     pfdata[17] = pfop->Px();	// px                 of this pfo
     pfdata[18] = pfop->Py();	// py                 of this pfo
     pfdata[19] = pfop->Pz();	// pz                 of this pfo
     pfdata[20] = gppid;        // pid of the grand parent
     pfdata[21] = pidpbp;       // pid of the primary break point
     pfdata[22] = vtx.X();	// vx                 of this pfo
     pfdata[23] = vtx.Y();	// vy                 of this pfo
     pfdata[24] = vtx.Z();	// vz                 of this pfo
     tupp  ->Fill(pfdata);
     TObjArray *childrenp = pfop->GetDaughterPFOsPtr();
     if (childrenp && childrenp->GetEntries() == 2) {
       S4PFObject *pfo1p = static_cast<S4PFObject *>(childrenp->At(0));
       static Int_t kNv0dat = 25;
       Double_t v0data[kNv0dat];
       v0data[ 0] = nev;		// event number
       v0data[ 1] = npj[0];		// # pfos     in jet1
       v0data[ 2] = npj[1];		// # pfos     in jet2
       v0data[ 3] = cosj[0];		// cos(theta) of jet1
       v0data[ 4] = cosj[1];		// cos(theta) of jet2
       v0data[ 5] = evis;		// Evis       of the final state
       v0data[ 6] = pt;			// Pt         of the final state
       v0data[ 7] = pl;			// Pl         of the final state
       v0data[ 8] = mass;		// inv. mass  of the final state
       v0data[ 9] = pfop->GetBreakPointPtr()->GetParticleID();	// pid     of this v0
       v0data[10] = pfop->GetQuality();				// quality of this v0
       v0data[11] = pfo1p->GetPosition().X();			// vx      of this v0
       v0data[12] = pfo1p->GetPosition().Y();			// vy      of this v0
       v0data[13] = pfo1p->GetPosition().Z();			// vz      of this v0
       v0data[14] = pfop->Px();					// px      of this v0
       v0data[15] = pfop->Py();					// py      of this v0
       v0data[16] = pfop->Pz();					// pz      of this v0
       v0data[17] = pfop->E();					// energy  of this v0
       v0data[18] = pfo1p->GetBreakPointPtr()->GetPosition().X();	// vx:true
       v0data[19] = pfo1p->GetBreakPointPtr()->GetPosition().Y();	// vy:true
       v0data[20] = pfo1p->GetBreakPointPtr()->GetPosition().Z();	// vz:true
       v0data[21] = pfop->GetBreakPointPtr()->GetFourMomentum().Px();	// px:true
       v0data[22] = pfop->GetBreakPointPtr()->GetFourMomentum().Py();	// py:true
       v0data[23] = pfop->GetBreakPointPtr()->GetFourMomentum().Pz();	// pz:true
       v0data[24] = pfop->GetBreakPointPtr()->GetFourMomentum().E();	// energy:true
       tupv0p->Fill(v0data);
     }
     pbpp->Lock();			// lock used primary BP
   }
   // -------------------------
   //  Undetected particles
   // -------------------------
   static Int_t kNahdat = 17;
   Double_t ahdata[kNahdat];
   ANL4DVector qah;
   ANL4DVector qahznu;
   ANL4DVector qahhnu;
   ANL4DVector qahisr;
   nextpbp.Reset();
   while ((pbpp = static_cast<S4BreakPoint *>(nextpbp()))) {
     Int_t j4tkid = pbpp->GetTrackID();
     MyPrimaryParticle *myppp  = MyPrimaryParticle::GetMyPPPtr(j4tkid);
     MyPrimaryParticle *mygppp = myppp  
                               ? (myppp->GetGrandParentPtr() ? myppp->GetGrandParentPtr() : myppp)
                               : 0;
     Double_t           gppid  = mygppp ? mygppp->GetParticleID() : 0;
     Double_t           agppid = TMath::Abs(gppid);
     Double_t           ppid   = myppp
	                       ? (myppp->GetParentPtr() ? myppp->GetParentPtr()->GetParticleID() : 0)
                               : 0;
     Double_t           appid  = TMath::Abs(ppid);
     if (!pbpp->IsLocked()) {		// select unused primary BP
       qah  += pbpp->GetFourMomentum();
       Double_t pid  = pbpp->GetParticleID();
       Double_t apid = TMath::Abs(pid);
       if (TMath::Abs(appid-23.)<0.1	// neutrinos from Z -> nu nu
           && (TMath::Abs(apid-12.)<0.5 || TMath::Abs(apid-14.)<0.5 || TMath::Abs(apid-16.)<0.5)) {
           qahznu  += pbpp->GetFourMomentum();
       } else if (TMath::Abs(agppid-25.)<0.1	// neutrinos from H
           && (TMath::Abs(apid-12.)<0.5 || TMath::Abs(apid-14.)<0.5 || TMath::Abs(apid-16.)<0.5)) {
           qahhnu  += pbpp->GetFourMomentum();
       } else if (TMath::Abs(agppid-22.)<0.1) {	// ISR photons
           qahisr  += pbpp->GetFourMomentum();
       }
       ahdata[ 0] = nev;		// event number
       ahdata[ 1] = npj[0];		// # pfos     in jet1
       ahdata[ 2] = npj[1];		// # pfos     in jet2
       ahdata[ 3] = cosj[0];		// cos(theta) of jet1
       ahdata[ 4] = cosj[1];		// cos(theta) of jet2
       ahdata[ 5] = evis;		// Evis       of the final state
       ahdata[ 6] = pt;			// Pt         of the final state
       ahdata[ 7] = pl;			// Pl         of the final state
       ahdata[ 8] = mass;		// inv. mass  of the final state
       ahdata[ 9] = pbpp->GetCharge();			// charge of this undetected particle
       ahdata[10] = pid;                                // pid    of this undetected particle
       ahdata[11] = gppid;				// parent id
       ahdata[12] = pbpp->GetFourMomentum().E(),	// energy of this undetected particle
       ahdata[13] = pbpp->GetFourMomentum().Px(),	// px     of this undetected particle
       ahdata[14] = pbpp->GetFourMomentum().Py(),	// py     of this undetected particle
       ahdata[15] = pbpp->GetFourMomentum().Pz();	// pz     of this undetected particle
       tupahp->Fill(ahdata);
     }
   }

   // -------------------------
   //  Store event information
   // -------------------------
   Double_t evdata[75];
   evdata[ 0] = nev;		// event number
   evdata[ 1] = npj[0];		// # pfos     in jet1
   evdata[ 2] = npj[1];		// # pfos     in jet2
   evdata[ 3] = cosj[0];	// cos(theta) of jet1
   evdata[ 4] = cosj[1];	// cos(theta) of jet2
   evdata[ 5] = evis;		// Evis       of the final state
   evdata[ 6] = pt;		// Pt         of the final state
   evdata[ 7] = pl;		// Pl         of the final state
   evdata[ 8] = mass;		// inv. mass  of the final state
   evdata[ 9] = Npfos;		// # pfos     in the final state
   evdata[10] = qem.E();	// E(EM)      of the final state
   evdata[11] = qhd.E();	// E(HD)      of the final state
   evdata[12] = qtk.E();	// E(TK)      of the final state
   evdata[13] = qembp.E();	// E(EM):bp   of the final state
   evdata[14] = qhdbp.E();	// E(HD):bp   of the final state
   evdata[15] = qtkbp.E();	// E(TK):bp   of the final state
   evdata[16] = qempbp.E();	// E(EM):pbp  of the final state
   evdata[17] = qhdpbp.E();	// E(HD):pbp  of the final state
   evdata[18] = qtkpbp.E();	// E(TK):pbp  of the final state
   evdata[19] = qah.E();	// E(AH):pbp  of the final state
   evdata[20] = qahznu.E();	// E(AH):pbp  of the final state Z->nu nu, if any
   evdata[21] = qahisr.E();	// E(AH):pbp  of the final state ISR photons
   evdata[22] = qisr.E();	// E(ISR)     of the final state
   evdata[23] = qisrbp.E();	// E(ISR):bp  of the final state
   evdata[24] = qisrpbp.E();	// E(ISR):pbp of the final state
   evdata[25] = qtot.Px();	// Px         of the final state
   evdata[26] = qtot.Py();	// Py         of the final state
   evdata[27] = qtot.Pz();	// Pz         of the final state
   evdata[28] = qisr.Px();	// Px(ISR)    of the final state
   evdata[29] = qisr.Py();	// Py(ISR)    of the final state
   evdata[30] = qisr.Pz();	// Pz(ISR)    of the final state
   evdata[31] = qah.Px();	// Px(AH):pbp of the final state
   evdata[32] = qah.Py();	// Py(AH):pbp of the final state
   evdata[33] = qah.Pz();	// Pz(AH):pbp of the final state
   evdata[34] = qahznu.Px();	// Px(AH):pbp of the final state Z->nu nu, if any
   evdata[35] = qahznu.Py();	// Py(AH):pbp of the final state Z->nu nu, if any
   evdata[36] = qahznu.Pz();	// Pz(AH):pbp of the final state Z->nu nu, if any
   evdata[37] = qahisr.Px();	// Px(AH):pbp of the final state ISR photons
   evdata[38] = qahisr.Py();	// Py(AH):pbp of the final state ISR photons
   evdata[39] = qahisr.Pz();	// Pz(AH):pbp of the final state ISR photons
   evdata[40] = qahhnu.E();	// E(AH):pbp  of the final state nu from H, if any
   evdata[41] = qahhnu.Px();	// Px(AH):pbp of the final state nu from H, if any
   evdata[42] = qahhnu.Py();	// Py(AH):pbp of the final state nu from H, if any
   evdata[43] = qahhnu.Pz();	// Pz(AH):pbp of the final state nu from H, if any
   evdata[44] = qem.Px();	// Px:em      of the final state
   evdata[45] = qem.Py();	// Py:em      of the final state
   evdata[46] = qem.Pz();	// Pz:em      of the final state
   evdata[47] = qhd.Px();	// Px:hd      of the final state
   evdata[48] = qhd.Py();	// Py:hd      of the final state
   evdata[49] = qhd.Pz();	// Pz:hd      of the final state
   evdata[50] = qdbltk.E();	// E:tk:dble  of the final state
   evdata[51] = qdbltk.Px();	// E:tk:dble  of the final state
   evdata[52] = qdbltk.Py();	// E:tk:dble  of the final state
   evdata[53] = qdbltk.Pz();	// E:tk:dble  of the final state
   evdata[54] = qdblem.E();	// E:em:dble  of the final state
   evdata[55] = qdblem.Px();	// E:em:dble  of the final state
   evdata[56] = qdblem.Py();	// E:em:dble  of the final state
   evdata[57] = qdblem.Pz();	// E:em:dble  of the final state
   evdata[58] = qdblhd.E();	// E:hd:dble  of the final state
   evdata[59] = qdblhd.Px();	// E:hd:dble  of the final state
   evdata[60] = qdblhd.Py();	// E:hd:dble  of the final state
   evdata[61] = qdblhd.Pz();	// E:hd:dble  of the final state
   evdata[62] = qdbltkab.E();	// E:tk:dble:ab  of the final state
   evdata[63] = qdbltkab.Px();	// E:tk:dble:ab  of the final state
   evdata[64] = qdbltkab.Py();	// E:tk:dble:ab  of the final state
   evdata[65] = qdbltkab.Pz();	// E:tk:dble:ab  of the final state
   evdata[66] = qtkmpc.E();	// E:tk:mpc      of the final state
   evdata[67] = qtkmpc.Px();	// E:tk:mpc      of the final state
   evdata[68] = qtkmpc.Py();	// E:tk:mpc      of the final state
   evdata[69] = qtkmpc.Pz();	// E:tk:mpc      of the final state
   evdata[70] = qdbltkmpc.E();	// E:tk:dble:mpc of the final state
   evdata[71] = qdbltkmpc.Px();	// E:tk:dble:mpc of the final state
   evdata[72] = qdbltkmpc.Py();	// E:tk:dble:mpc of the final state
   evdata[73] = qdbltkmpc.Pz();	// E:tk:dble:mpc of the final state
   evdata[74] = qtkbpmpc.E();	// E(TK):bp:mpc  of the final state
   tupep->Fill(evdata);
#if 1
   cerr << " Npfos =  " << Npfos << endl
        << " Eem = " << qem.E() << " Eembp = " << qembp.E() << endl
        << " Ehd = " << qhd.E() << " Ehdbp = " << qhdbp.E() << endl
        << " Etk = " << qtk.E() << " Etkbp = " << qtkbp.E() << endl
	<< " Epfo = " << qem.E() + qhd.E() + qtk.E() << " Ebp = " << qembp.E() + qhdbp.E() + qtkbp.E() << endl
        << " Epbp = " << qempbp.E() + qhdpbp.E() + qtkpbp.E() 
	<< " Eah = " << qah.E() << " EahZnu = " << qahznu.E() << " EahHnu = " << qahhnu.E() << " EahISR = " << qahisr.E() << endl
	<< " Epbp + Eah = " << qempbp.E() + qhdpbp.E() + qtkpbp.E() + qah.E() << endl
	<< " npj1 = " << npj[0] << " npj2 = " << npj[1] << " cosj1 = " << cosj[0] << " cosj2 = " << cosj[1] << endl
	<< " Pt = " << pt << " Pl = " << pl << " Evis = " << evis << " Mass = " << mass << endl;
   cerr << " (E, Px, Py, Pz):tot = (" << qtot.E() << ", " << qtot.Px() << ", " << qtot.Py() << ", " << qtot.Pz() << ")" << endl
        << " (E, Px, Py, Pz):isr = (" << qisr.E() << ", " << qisr.Px() << ", " << qisr.Py() << ", " << qisr.Pz() << ")" << endl
        << " Mjj(w isr)   = " << qtot.GetMass() << endl
        << " Mjj(w/o isr) = " << (qtot-qisr).GetMass() << endl
        << " Mjj(w/o isr + ah) = " << (qtot-qisr+qah-qahznu-qahisr).GetMass() << endl;
#endif

   // -------------------------
   //  That's it, folks:^)
   // -------------------------
   last->cd();
}

TObjArray & MyPFOAnalysis::GetPrimaryBPs()
{
   if (!fPrimaryBPs.GetEntries()) {
     S4PrimaryVertex *pvp = static_cast<S4PrimaryVertex *>(
		              static_cast<S4GENMCTruthBuf *>(
                                gJSF->FindModule("S4GENMCTruth")->EventBuf()
			      )->GetOutputBufPtr()->At(0)
                            );
     TObjArray *pplist = pvp->GetParticles();
     MyPrimaryParticle::Initialize(pplist);

     TIter next(pplist);
     S4PrimaryParticle *ppp;
     while ((ppp = static_cast<S4PrimaryParticle *>(next()))) {
       if (ppp->GetTrackID() >= 0 && !ppp->GetDaughters()->GetEntries()) { // primary with no daughters
         Int_t j4tkid = ppp->GetTrackID();
         S4BreakPoint *bpp = S4BreakPoint::GetBreakPointPtr(j4tkid);
         if (bpp) {
           bpp->SetBit(BIT(20));		// flag this as a primary BP
           fPrimaryBPs.Add(bpp);		// store this in the primary BP TObjArray
         } else {
           cerr << ">>>>>> Error in MyPFOAnalysis::GetPrimaryBPs() " << endl
                << " bpp = 0! unexpected! abort.                   " << endl;
           abort();
         }
       }
     }
   }
   return fPrimaryBPs;
}

S4JetSet * MyPFOAnalysis::GetJetSet(Int_t njets)
{
   S4JetMakerBuf *jmbp = static_cast<S4JetMakerBuf *>( gJSF->FindModule("S4JetMaker")->EventBuf());
   return jmbp->GetJetSetPtr(njets);
}
