//*************************************************************************
//* ---------------------
//* S4BabyVectorHit Class
//* ---------------------
//* (Description)
//*	Class to store "S4BabyHitWire"s for a single cell.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	S4BabyVectorHit		: Array containing "S4BabyHitWire"s
//* (Update Record)
//*	2002/08/28  R.Kuboshima
//*************************************************************************

#include <iostream>
#include <iomanip>
#include "TMath.h"
#include "S4BabyVectorHit.h"

Int_t S4BabyVectorHit::fSortBy =  0;           // (0,1) = (chi2,time)

ClassImp(S4BabyVectorHit)

//=====================================================================
//* constructor -------------------------------------------------------

S4BabyVectorHit::S4BabyVectorHit() 
      : TObjArray(5),
	fKilled(kFALSE), fTrackNumber(0), fCellNumber(0),
	fTrackPtr(0), fNExpectedCellHitsCut(8),
	fDone(kFALSE), fTrackType(0), fChi2(0.),
        fa(0.), fb(0.), fEaa(0.), fEba(0.), fEbb(0.)
{
	SetOwner();
}

S4BabyVectorHit::S4BabyVectorHit(const S4BabyVectorHit &t) 
      : TObjArray(5), TAttLockable(t),
	fKilled(t.fKilled), 
	fTrackNumber(t.fTrackNumber), 
	fCellNumber(t.fCellNumber),
        fTrackPtr(t.fTrackPtr), 
	fNExpectedCellHitsCut(t.fNExpectedCellHitsCut),
        fDone(t.fDone), fTrackType(t.fTrackType), fChi2(t.fChi2),
        fa(t.fa), fb(t.fb), fEaa(t.fEaa), fEba(t.fEba), fEbb(t.fEbb)
{
       	TIter next(&t);
        S4BabyTrackHit *hitp;
	while ((hitp = (S4BabyTrackHit *)next())) {
		Add(new S4BabyTrackHit(*hitp));
	}
	SetOwner();
}


//=====================================================================
//* destructor --------------------------------------------------------

S4BabyVectorHit::~S4BabyVectorHit()
{
}

//=====================================================================
//* Compare -----------------------------------------------------------

Int_t S4BabyVectorHit::Compare(const TObject *obj) const
{
	Int_t cell  = fCellNumber;
	Int_t cello = ((S4BabyVectorHit *)obj)->fCellNumber;
   	if (cell < cello) return -1;
   	else if (cell > cello) return 1;
   	else return 0;
}

//=====================================================================
//* Append ------------------------------------------------------------

Int_t S4BabyVectorHit::Append(S4BabyTrackHit *trackhit)
{	
	Int_t track  = trackhit->GetTrackID();
	Int_t cellno = trackhit->GetExactHitPtr()->GetCellNumber(); 
	
	// Return value = -1:	new trackhit.
	//		=  0:	successfully appended.
	//		=  1:	new cell.
	//		=  2:	new track.
	
	if (GetEntries() == 0) {
		fTrackNumber = track;
		fCellNumber  = cellno;
		trackhit->fVectorHitPtr = this;
                Add(trackhit);
                return 0;
	} else {
		if (track == fTrackNumber) {
			if (cellno == fCellNumber) {
				trackhit->fVectorHitPtr = this;
                        	Add(trackhit);
                        	return 0;
                        } else {
                        	return 1;
                        }
                } else {
                        return 2;
                }
        }
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void S4BabyVectorHit::DebugPrint(const Char_t *opt)
{
	std::cerr 	<< "(track,cell) = " 
		<< "(" << fTrackNumber << "," << fCellNumber << ")" 
		<< " NEntries = " << GetEntries() << std::endl;
		
#if 0
	TIter next(this);
	S4BabyTrackHit *trackhit;
	Int_t n = 0;
	while ((trackhit = (S4BabyTrackHit *)next())) {
		std::cerr << std::setw(3) << ++n << " ";
		trackhit->DebugPrint(opt);
	}
#endif
}

//=====================================================================
//* KillHits ----------------------------------------------------------

void S4BabyVectorHit::KillHits() 
{
	if (GetEntries() > fNExpectedCellHitsCut) {
		SetBit(kELooping);
	}
		
	fKilled = kTRUE;
}	
	
//=====================================================================
//* Fit ----------------------------------------------------------

void S4BabyVectorHit::Fit()
{
        Double_t det, x, y, rs2;
        Double_t syy = 0.;
        Double_t sy  = 0.;
        Double_t s   = 0.;
        Double_t sxy = 0.;
        Double_t sx  = 0.;
        TIter next(this);
        U4CDCTrackHit *tkhitp;
        while ((tkhitp = (U4CDCTrackHit *)next())) {
                U4CDCTrackHit  &tkhit = *tkhitp;
                U4CDCHit       &hit   = *tkhit.GetHitPtr();
                if (tkhit.IsLocked()) continue;
                rs2  = 1./TMath::Power(hit.DXdrift(),2);
                x    = hit.Xwire() + (2*tkhit.GetLeftRight() - 1) * hit.Xdrift();
                y    = hit.Ywire();
                syy += rs2*y*y;
                sy  += rs2*y;
                s   += rs2;
                sxy += rs2*x*y;
                sx  += rs2*x;
        }
        det  = syy*s -sy*sy;
        fa   = (  s*sxy -  sy*sx)/det;
        fb   = (-sy*sxy + syy*sx)/det;
        fEaa =   s/det;
        fEba = -sy/det;
        fEbb = syy/det;

        fChi2 = 0.;
        next.Reset();
        while ((tkhitp = (U4CDCTrackHit *)next())) {
                U4CDCTrackHit  &tkhit = *tkhitp;
                U4CDCHit       &hit   = *tkhit.GetHitPtr();
                x = hit.Xwire() + (2*tkhit.GetLeftRight() - 1) * hit.Xdrift();
                y = hit.Ywire();
                tkhit.SetResidual(x - fa*y - fb);
                fChi2 += TMath::Power(tkhit.GetResidual()/hit.DXdrift(),2);
        }

}

