#ifndef __U4CDCRAWDATA__
#define __U4CDCRAWDATA__

//*************************************************************************
//* ----------------
//* U4CDCRawData Class
//* ----------------
//* (Description)
//*	This set of classes reads U4CDC raw data for Lorentz Angle 
//*	measurments and unpacks them into root objects.
//*	When a root file is opened in the read/write mode, the root
//*	objects are stored in it.  If the root file is read-only,
//*	the data are obtained from it.
//*	The package contains the following classes:
//*	The following parameters can be specified in "jsf.conf". 
//*
//*	    U4CDCRawData.RFNFormat:       printf format to construct file name
//*	    U4CDCRawData.DoPlots:         (TRUE/FASLE)
//* (Requires)
//*	TObject:
//*	TClonesArray:
//*	U4CDCUniversalRecordHeader:
//*	U4CDCRawADCDatum:
//* (Provides)
//*	U4CDCHardwareMap		: hardware map
//*	U4CDCBeginRunBuf		: begin run buffer
//*	U4CDCRawDataBuf		: raw data buffer
//*	U4CDCEndRunBuf		: end run buffer
//*	U4CDCRawData		: steering class
//* (Update Record)
//*	1999/06/02  K.Fujii	Original version.
//*	2000/05/17  K.Fujii	Implementation for new format (not tested)
//*	2001/05/27  K.Fujii	Modification for new branch scheme
//*				started from ROOT v.3.01.
//*	2001/09/14  K.Fujii	Use automatic schema evolution.
//*************************************************************************

#include "TObject.h"
#include "TArrayI.h"
#include "TObjString.h"
#include "TClonesArray.h"
#include "TDatime.h"

#include "U4VTypeDef.h"
#include "U4CDCUniversalRecordHeader.h"
#include "U4CDCRawGPIBDatum.h"
#include "U4CDCRawScalerDatum.h"
#include "U4CDCRawADCDatum.h"
#include "TDynArray3.h"

class U4CDCHardwareMap : public TDynArray3 {
public:
  U4CDCHardwareMap() {}
  virtual ~U4CDCHardwareMap() {}

  ClassDef(U4CDCHardwareMap, 1) // U4CDC HardwareMap
};

class U4CDCChannel : public TObject {
friend class U4CDCBeginRunBuf;
public:
  U4CDCChannel(Int_t id = 0, Int_t ch = 0) : fID(id), fDetChNo(ch) {}
  virtual ~U4CDCChannel() {}
private:
  inline Int_t GetID   () const { return fID;      }
  inline Int_t GetDetCh() const { return fDetChNo; } 

private:
  Int_t fID;		// detector ID
  Int_t fDetChNo;	// detector channel No

  ClassDef(U4CDCChannel, 1) // FADC channel map
};

// ----------------
// Begin Run Buffer
// ----------------

class U4CDCRawData;
class U4CDCBeginRunBuf : public TObject {
public:
  U4CDCBeginRunBuf();
  virtual ~U4CDCBeginRunBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(BRR_t data[]);
  
  inline U4CDCUniversalRecordHeader *GetHeader  ()       { return &fHeader;   }
  inline Int_t                       GetDateTime() const { return  fDateTime; }
  inline TObjString                 *GetComment ()       { return &fComment;  }

  void   DebugPrint (const Char_t *opt = "")  {}
  void   EChannelToDChannel(Int_t cr, Int_t st, Int_t ch, 
  			    Int_t &id, Int_t &dch);

private:
  void   UnpackDateTime(BRR_t data[]);
  void   UnpackHardwareMap(BRR_t data[]);
  void   UnpackThresholdData(BRR_t data[]);


private:
  U4CDCUniversalRecordHeader  fHeader;	// Header
  Int_t		   	    fDateTime ;	// Date and time
  U4CDCHardwareMap	    fMap;	// (sta#,ch#) --> detector map
  TArrayI          	    fThresh;	// threshold data 
  TObjString		    fComment;	// Comment string

  ClassDef(U4CDCBeginRunBuf, 1) // U4CDC BeginRun record
};

// ---------------
// Raw Data Buffer
// ---------------

class U4CDCRawDataBuf : public JSFEventBuf {
private:
  U4CDCUniversalRecordHeader  fHeader; 		// Universal Record Header
  Len_t		    	    fNGPIBData;		// No. GPIB data units
  Len_t		    	    fNScalerData;	// No. Scaler data units
  Len_t		    	    fNADCData;		// No. ADC data units
  TClonesArray		   *fGPIBData;		// Array of GPIB Data
  TClonesArray		   *fScalerData;	// Array of Scaler Data
  TClonesArray		   *fADCData;		// Array of ADC Data
  Int_t 		    fNonFADCIndex; 	// index of 1st nonFADC datum
public:
  U4CDCRawDataBuf(const char *name="U4CDCRawDataBuf",
		const char *title="U4CDC raw data", U4CDCRawData *mod=0);
  U4CDCRawDataBuf(U4CDCRawData *mod, const char *name="U4CDCRawDataBuf",
		const char *title="U4CDC raw data");
  virtual ~U4CDCRawDataBuf();

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ADC_t data[]);
  inline U4CDCUniversalRecordHeader * GetHeader()   { return &fHeader;      }
  inline Len_t		  GetNGPIBData   () const { return  fNGPIBData;   }
  inline Len_t		  GetNScalerData () const { return  fNScalerData; }
  inline Len_t		  GetNADCData    () const { return  fNADCData;    }
  inline TClonesArray	* GetGPIBData    () const { return  fGPIBData;    }
  inline TClonesArray	* GetScalerData  () const { return  fScalerData;  }
  inline TClonesArray   * GetADCData     () const { return  fADCData;     }
  inline Int_t 		  GetNonFADCIndex() const { return  fNonFADCIndex;}

  ClassDef(U4CDCRawDataBuf, 3) // U4CDC raw data buffer
};

// --------------
// End Run Buffer
// --------------

class U4CDCEndRunBuf : public TObject {
private:
  U4CDCUniversalRecordHeader  fHeader;	// Header 
  Int_t		   	    fDateTime ;	// Date and time
  TObjString		    fComment;	// Comment string
public:
  U4CDCEndRunBuf() {}
  virtual ~U4CDCEndRunBuf() {}

  Bool_t ReadFile(FILE *fd);
  void   Unpack(ERR_t data[]);
  
  inline U4CDCUniversalRecordHeader * GetHeader  ()       { return &fHeader;   }
  inline Int_t                        GetDateTime() const { return  fDateTime; }
  inline TObjString                 * GetComment ()       { return &fComment;  }
         void                         DebugPrint (const Char_t *opt = "") {}

  ClassDef(U4CDCEndRunBuf, 1) // U4CDC EndRun record
};

// -----------------
// Raw Data Analysis
// -----------------

class U4CDCRawData : public JSFModule {
private:
   U4CDCBeginRunBuf    *fBeginBuf;	//  Begin Run Record
   Char_t   	     *fRFNFormat;	//! Input file name format
   FILE     	     *fd;		//! File descriptor for the raw data

public:
   U4CDCRawData(const Char_t *name="U4CDCRawData",
	      const Char_t *title="Get U4CDC Raw data");
   virtual ~U4CDCRawData();

   void MakeBranch(TTree *tree);
   void SetBranch(TTree *tree);

   Bool_t BeginRun(Int_t nrun);
   Bool_t Process (Int_t ev);
   void   DebugPrint(const Char_t *opt = "") {}

   inline U4CDCBeginRunBuf    * GetBeginRunBuf() const { return fBeginBuf; }

   ClassDef(U4CDCRawData, 1) // Read U4CDC raw data analysis
};

#endif


