//*************************************************************************
//* ==============
//* U4VTrack Class
//* ==============
//* (Description)
//*    Base class to all kinds of uranus charged tracks.
//* (Requires)
//*    TObjArray
//* (Provides)
//*    U4VTrack    : Array containing "U4VTrackHit"s
//* (Update Record)
//*    2004/08/08  K.Fujii         Original version.
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "JSFUtil.h"
#include "JSFHelicalTrack.h"

#include "U4VTrack.h"
#include "U4VTrackHit.h"
#include "U4VTrackMaker.h"
#include "U4IPKalHit.h"

#include "TKalTrackState.h"
#include "TKalTrackSite.h"
#include "TKalTrack.h"

Int_t  U4VTrack::fSortBy =  0;           // (0,1) = (chi2,time)

ClassImp(U4VTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4VTrack::U4VTrack(U4VTrackMaker* trkm)
        : TObjArray(200), 
          fTrackNo(0), 
          fDone(kFALSE), 
          fTrackType(0), 
          fChi2(0.), 
          fNDF(0),
          fT0(0.), 
          fIPConstraint(kFALSE),
          fHelicalTrackPtr(0),
          fTrackMakerPtr(trkm),
          fSeedTrackPtr(0)
{
   SetOwner();
}

U4VTrack::U4VTrack(const U4VTrack &tk)
        : TObjArray(tk.GetEntries()),
          fTrackNo(tk.fTrackNo),
          fDone(tk.fDone),
          fTrackType(tk.fTrackType),
          fChi2(tk.fChi2),
          fNDF(tk.fNDF),
          fT0(tk.fT0),
          fIPConstraint(tk.fIPConstraint),
          fKalTrack(tk.fKalTrack),
          fHelicalTrackPtr(tk.fHelicalTrackPtr),
          fTrackMakerPtr(tk.fTrackMakerPtr),
          fSeedTrackPtr(tk.fSeedTrackPtr)
{
   // Must to do deep copy for U4VTrackHit's in the array
   // in the derived class 
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4VTrack::~U4VTrack()
{
   if (fHelicalTrackPtr) delete fHelicalTrackPtr;
}


//=====================================================================
//* IsSortable --------------------------------------------------------

Bool_t U4VTrack::IsSortable() const
{ 
   return kTRUE;
}

//=====================================================================
//* Compare -----------------------------------------------------------

Int_t U4VTrack::Compare(const TObject *obj) const
{
   U4VTrack *tk     = (U4VTrack *)this;
   U4VTrack *tko    = (U4VTrack *)obj;
   switch (GetSortBy()) {
      case 1: { // sort by time: faster one first:-)
               Double_t time  = tk ->GetT0();
               Double_t timeo = tko->GetT0();
               return (time < timeo ? -1 : (time > timeo ? +1 : 0));
      }
      case 2: { // sort by track number
              Int_t track  = tk ->GetTrackNumber();
              Int_t tracko = tko->GetTrackNumber();
               return (track < tracko ? -1 : (track > tracko ? +1 : 0));
      }
      default: { // sort by chi2: better one first:-)
               Double_t chi2  = tk ->GetChi2();
               Double_t chi2o = tko->GetChi2();
               return (chi2 < chi2o ? -1 : (chi2 > chi2o ? +1 : 0));
      }
   }
}

//=====================================================================
//* SetSortBy ---------------------------------------------------------

void U4VTrack::SetSortBy(Char_t *opt)
{
   switch (opt[0]) {
      case 'T':       // sort by "Time"
         fSortBy       = 1;
         break;
      case 'S':       // sort by "SerialNumber"
         fSortBy       = 2;
         break;
      default:        // sort by "Chi2"
         fSortBy       = 0;
         break;
   }
}

void U4VTrack::SetSortBy(Int_t key)
{
   fSortBy = key;
}

//=====================================================================
//* Update ------------------------------------------------------------

void U4VTrack::Update(const Char_t *mode)
{
   switch (mode[0]) {
      case 'H' : // Helix
#ifdef __DEBUG__
         std::cerr << " HelixFit is called --------------------------- " << std::endl;
#endif
         HelixFit();
         break;
      case 'K' : // Kalman
#ifdef __DEBUG__
         std::cerr << " KalmanFilter is called ----------------------- " << std::endl;
#endif
         KalmanFilter();
         break;
      default :
         std::cerr << " U4VTrack::Update:"
                   << " Wrong Option! Use : " << std::endl;
         std::cerr << "    Helix " << std::endl;
         std::cerr << "    Kalman " << std::endl;
   }
}

//=====================================================================
//* Add ---------------------------------------------------------------

void U4VTrack::Add(TObject *obj)
{
	TObjArray::Add(obj);
        U4VTrackHit *vhitp = (U4VTrackHit *)obj;
        if (!vhitp->IsLocked()) fDone = kFALSE;
}

//=====================================================================
//* HelixFit ----------------------------------------------------------

void U4VTrack::HelixFit()
{
   if (!fDone) KalmanFilter();

   TKalTrackSite &site = dynamic_cast<TKalTrackSite &>(fKalTrack.GetCurSite());
   TKalTrackState ahel = TKalTrackState(site.GetCurState(), site);
                                        
   TKalMatrix     Chel(kSdim,kSdim);
   Int_t          ndfhel;

   (*(TKalTrackSite *)fKalTrack.At(0)).Lock(); // don't use dummy site

   fChi2 = fKalTrack.FitToHelix(ahel,Chel,ndfhel);
   fNDF  = ndfhel;
   fT0   = ahel(5,0); // [nsec]

   SetJSFHelicalTrack(site, ahel, Chel);

#ifdef __DEBUG__
   static TH1D *hChi2  = 0;
   static TH1D *hNdf   = 0;
   static TH1D *hCL    = 0;
   static TH1D *hT0    = 0;
   static TH2D *hNdfT0 = 0;
   
   if (!hChi2)  hChi2  = new TH1D("hChi2h","",200, 0., 200.);
   if (!hNdf)   hNdf   = new TH1D("hNdfh","",200, 0., 200.);
   if (!hCL)    hCL    = new TH1D("hCLh","",50, 0., 1.);
   if (!hT0)    hT0    = new TH1D("hT0h","",200, -50., 50.);
   if (!hNdfT0) hNdfT0 = new TH2D("hNdfT0h","", 200, 0., 200, 200, -50., 50.);

   hNdf->Fill(fNDF);
   hChi2->Fill(fChi2);
   hCL->Fill(fHelicalTrackPtr->GetCL());
   hT0->Fill(fT0);
   hNdfT0->Fill(fNDF, fT0, 1.);
#endif
}

//=====================================================================
//* SetJSFHelicalTrack ------------------------------------------------

void U4VTrack::SetJSFHelicalTrack(const TKalTrackSite  &site,
                                  const TKalTrackState &a,
                                  const TKalMatrix     &cov)
{
   // Save the fit result ==============================================

   Float_t pivot[3];
   pivot[0] = site.GetPivot().X();
   pivot[1] = site.GetPivot().Y();
   pivot[2] = site.GetPivot().Z();

   Float_t stv[6];
   stv[0] = a(0,0);
   stv[1] = a(1,0);
   stv[2] = a(2,0);
   stv[3] = a(3,0);
   stv[4] = a(4,0);
   stv[5] = a(5,0);

   Double_t error[15];
   Int_t    n = 0;
   for (Int_t i=0; i<5; i++) {
      for (Int_t j=0; j<=i; j++) {
         error[n++] = cov(j,i);
      }
   }
   Double_t chi2 = fChi2;
   Int_t    ndf  = fNDF;
   Double_t cl   = JSFUtil::ConfidenceLevel(ndf,chi2);
   
   if (fHelicalTrackPtr) delete fHelicalTrackPtr;
   fHelicalTrackPtr = new JSFHelicalTrack(&stv[0], &pivot[0],
                                          ndf, chi2, cl, error);
   fHelicalTrackPtr->SetBfield(site.GetHit().GetBfield());
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4VTrack::DebugPrint(const Char_t *opt)
{
   std::cerr << "TrackNumber = " << fTrackNo
             << "  NEntries = "  << GetEntries() << std::endl;
   if (strstr(opt,"Detailed")) {
      if (!fDone) Update();
      TKalTrackState a = fKalTrack.GetCurSite().GetCurState();

      std::cerr.setf(std::ios::right);
      std::cerr.setf(std::ios::fixed);
      std::cerr << " Type = " << std::setw(2) << fTrackType
                << " chi2 = " << std::setw(6) << std::setprecision(3) << fChi2
                << " ndf = " << std::setw(5) << fNDF
                << " t0 = " << std::setw(6) << std::setprecision(3) << fT0
                << " a = (" << std::setw(6) << std::setprecision(3) << a(0,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(1,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(2,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(3,0)
                << ","      << std::setw(6) << std::setprecision(3) << a(4,0)
                << ")"
                << std::endl;
   }

   TIter next(this);
   U4VTrackHit *vhit;
   while ((vhit = (U4VTrackHit *)next())) {
      vhit->DebugPrint(opt);
   }
}

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4VTrack::KalmanFilter(Bool_t dir)
{
   if (fDone) {
      TKalTrackSite  &site  = dynamic_cast<TKalTrackSite&>(fKalTrack.GetCurSite());
      TKalTrackState &state = dynamic_cast<TKalTrackState&>(site.GetCurState());
      Bool_t self = fSeedTrackPtr ? kFALSE : kTRUE;
      SetChi2(fKalTrack.GetChi2());
      SetNDF (fKalTrack.GetNDF(self));
      SetT0  (state(5,0));
      SetJSFHelicalTrack(site, state, state.GetCovMat());
      return;
   }
   fKalTrack.SetOwner();
   TObjArray kalhits(GetEntries());
   if (!fDone) MakeKalHits(kalhits);
   TKalTrackSite *firstsitep = GetFirstSitePtr(dir, &kalhits);
   if (fDone) {
      kalhits.Delete();
      delete firstsitep;
      return;
   }
   fDone = kTRUE;
   if (!firstsitep) {
      kalhits.Delete();
      Lock();
      return;
   }
   fKalTrack.Add(firstsitep);

   if (IsIPConstraint()) {
      if (dir == kIterForward) kalhits.Add(U4IPKalHit::Create());
      else                     kalhits.AddFirst(U4IPKalHit::Create());
   }
   TIter next(&kalhits, dir);
   TVTrackHit *hitp = 0;
   TKalTrackSite *sitep = 0;
   while ((hitp = static_cast<TVTrackHit *>(next()))) {
      sitep = new TKalTrackSite(*hitp);
      sitep->SetOwner();
      sitep->SetHitOwner();
      if (!fKalTrack.AddAndFilter(*sitep)) {
#ifdef __DEBUG__
         std::cerr << " site discarded!" << std::endl;
#endif
         delete sitep;
         sitep = 0;
      }
   }
#if 0
   if (sitep) {
      TKalTrackState &state  = dynamic_cast<TKalTrackState &>(sitep->GetCurState());
      Bool_t self = fSeedTrackPtr ? kFALSE : kTRUE;
      SetChi2(fKalTrack.GetChi2());
      SetNDF (fKalTrack.GetNDF(self));
      SetT0  (state(5, 0));
      SetJSFHelicalTrack(*sitep, state, state.GetCovMat());
   }
   if (!GetHelicalTrackPtr()) Lock();
#else
   if (fKalTrack.GetEntries() > 1) {
      TKalTrackSite  &site  = dynamic_cast<TKalTrackSite &>(fKalTrack.GetCurSite());
      TKalTrackState &state = dynamic_cast<TKalTrackState &>(site.GetCurState());
      SetChi2(fKalTrack.GetChi2());
      SetNDF (fKalTrack.GetNDF(!fSeedTrackPtr));
      SetT0  (state(5, 0));
      SetJSFHelicalTrack(site, state, state.GetCovMat());
   } else Lock();
#endif
}

//=====================================================================
//* SetNDF ------------------------------------------------------------

void U4VTrack::SetNDF(Int_t i)
{
   fNDF = !fSeedTrackPtr ?  i : i + fSeedTrackPtr->GetNDF();
}

//=====================================================================
//* SetChi2 -----------------------------------------------------------

void U4VTrack::SetChi2(Double_t d)
{
   fChi2 = !fSeedTrackPtr ?  d : d + fSeedTrackPtr->GetChi2();
}

