//*************************************************************************
//* ---------------------
//* U4VTXTrack Class
//* ---------------------
//* (Description)
//*	Class to store "U4VTXHit"s for a single track.
//* (Requires)
//*	TObjArray
//* (Provides)
//*	U4VTXTrack		: Array containing "U4VTXHit"s
//* (Update Record)
//*     2004/08/26   A.Yamaguchi   Original Version. 
//*************************************************************************

#include <iostream>
#include <iomanip>

#include "U4VTXTrack.h"
#include "U4VTXKalHit.h"
#include "U4VTXTrackHit.h"
#include "U4VTXMeasLayer.h"
#include "TKalTrackSite.h"
using namespace std;
//#define __DEBUG__

ClassImp(U4VTXTrack)

//=====================================================================
//* constructor -------------------------------------------------------

U4VTXTrack::U4VTXTrack()
{
   SetOwner();
}

U4VTXTrack::U4VTXTrack(U4VTXTrackMaker* trkm)
         : U4VTrack(trkm)
{
   SetOwner();
}

U4VTXTrack::U4VTXTrack(const U4VTXTrack &tk)
         : U4VTrack(tk)
{
   TIter next(&tk);
   U4VTXTrackHit *thitp = 0;
   while ((thitp = (U4VTXTrackHit *)next())) {
      Add(new U4VTXTrackHit(*thitp));
   }
   SetOwner();
}

//=====================================================================
//* destructor --------------------------------------------------------

U4VTXTrack::~U4VTXTrack()
{
}

//=====================================================================
//* MakeKalHits -------------------------------------------------------

void U4VTXTrack::MakeKalHits(TObjArray &kalhits)
{
   // -- Put track hits in a TObjArray -- //
   TIter nextthit(this);
   U4VTXTrackHit  *trackhitp = 0;

   while ((trackhitp = dynamic_cast<U4VTXTrackHit *>(nextthit()))) {
      if (trackhitp->IsLocked()) continue;
      U4VTXHit *hitp = trackhitp->GetHitPtr();

      Int_t    layerno = hitp->GetLayerID();
      Double_t bfield  = dynamic_cast<U4VTXKalDetector *>(GetTrackMakerPtr()->GetDetectorPtr())->GetBfield(layerno);

      Double_t meas [2];
      Double_t dmeas[2];
      meas  [0] = hitp->GetRPhi();
      meas  [1] = hitp->GetZ();
      dmeas [0] = hitp->GetDRPhi();
      dmeas [1] = hitp->GetDZ();

      const U4VTXMeasLayer &ms = *dynamic_cast<const U4VTXKalDetector *>
              (GetTrackMakerPtr()->GetDetectorPtr())->GetMeasLayerPtr(layerno);
      kalhits.Add(new U4VTXKalHit(*trackhitp, ms, meas, dmeas, bfield)); 

#ifdef __DEBUG__
      cerr.setf(ios::right);
      cerr.setf(ios::fixed);

      cerr << setw(4) << kalhits.GetEntries()
           << " rp = " 
           << setw(7) << setprecision(3) << ms.GetR() 
           << " "
           << "(layr) = ("
           << setw(2) << layerno << ")"
           << " "
           << " (rphi, z) = (" 
           << setw(8) << setprecision(4) << meas [0] << " +/-"
           << setw(8) << setprecision(4) << dmeas[0] << ","
           << setw(8) << setprecision(4) << meas [1] << " +/-"
           << setw(8) << setprecision(4) << dmeas[1] << ")"
           << endl;
      cerr << setprecision(7);

      cerr.unsetf(ios::right);
      cerr.unsetf(ios::fixed);
#endif
   }
}

//=====================================================================
//* GetFirstSitePtr ---------------------------------------------------

TKalTrackSite *U4VTXTrack::GetFirstSitePtr(Bool_t, TObjArray *)
{
   U4VTrack *stkp = GetSeedTrackPtr();
   if (!stkp) {
      // Self tracking mode that does not assume any seed track
      // Not implemented yet. Try following the TPC example
      // In that case, you should use dir and hitsp supplied as args
      cerr << " >>>>> U4VTXTrack::GetFirstSitePtr >>>>>>>>> " << endl
           << "  Supply  a seed track, since self tracking " << endl
           << "  is not yet implemented                    " << endl;
      return 0;
   }
   TKalTrackSite *sitep = new TKalTrackSite(dynamic_cast<TKalTrackSite &>
                                     (stkp->GetKalTrack().GetCurSite()));
   sitep->SetHitOwner(kFALSE);
   return sitep;
}

//=====================================================================
//* DebugPrint --------------------------------------------------------

void U4VTXTrack::DebugPrint(const Char_t *opt)
{
   cerr << "U4VTXTrack::DebugPrint: NEntries = "  << GetEntries() << endl;
   TString opts = opt;
   if (opts.Contains("Detailed")) {
      if (!IsDone()) Update();
      TKalTrackState a = GetKalTrack().GetCurSite().GetCurState();

      cerr.setf(ios::right);
      cerr.setf(ios::fixed);
      cerr << " Type = " << setw(2) << GetTrackType()
           << " chi2 = " << setw(6) << setprecision(3) << GetChi2()
           << " ndf = " << setw(5) << GetNDF()
           << " t0 = " << setw(6) << setprecision(3) << GetT0()
           << " a = (" << setw(6) << setprecision(3) << a(0,0)
           << ","      << setw(6) << setprecision(3) << a(1,0)
           << ","      << setw(6) << setprecision(3) << a(2,0)
           << ","      << setw(6) << setprecision(3) << a(3,0)
           << ","      << setw(6) << setprecision(3) << a(4,0)
           << ")"
           << endl;
   }
}

//=====================================================================
//* GetNhis --------------------------------------------------------

Int_t U4VTXTrack::GetNhits()
{
   Int_t nhits = 0;
   TIter next(this);
   U4VTXTrackHit *hitp = 0;
   while ((hitp = dynamic_cast<U4VTXTrackHit *>(next()))) {
      if (!hitp->IsLocked()) {
         nhits++;
      }
   }
   return nhits;
} 

//=====================================================================
//* KalmanFilter ------------------------------------------------------

void U4VTXTrack::KalmanFilter(Bool_t dir)
{
#ifdef __DEBUG__
   cerr << "U4VTXTrack::KalmanFilter is called..." << endl;
#endif
   SetTrackType(2);
   U4VTrack::KalmanFilter(dir);
}
